clc;

% number of modes
modes = 8;

% number of data points per mode, and total
points_per_mode = 100;
N = modes * points_per_mode;


% create / sample points?
create_points = 1;
sample_points = 1;

% initial number of components
K = 50;

% overlap threshold
tau = 0.55;

% expansion factor lambda [0 = default EM, 1 = expand only]
lambda = .2;

% show fig every n_it iterations
n_it = 1;

% maximum number of iterations
max_it = 400;

% dimensionality
D = 2;

% prior standard deviation
sigma0 = 0.00;
sigma_init = 0.02;

%% create random points
if(create_points == 1)

	% points per node
	nptspm = N /  modes;

	x = []; y = [];
	mode_var = 0.1 * rand(modes,1);
	mode_ctr_x = rand(modes,1);
	mode_ctr_y = rand(modes,1);
	mode_ctr = [mode_ctr_x mode_ctr_y];

	for i = 1:modes
		new_x =  mode_ctr_x(i) + mode_var(i) * randn(nptspm,1);
		x = [x; new_x];
		new_y = mode_ctr_y(i) + mode_var(i) * randn(nptspm,1);
		y = [y; new_y];
	end
	data = [x y];
end

%% sample points
if create_points || sample_points,
    randp = randperm(N);
    samples = data(randp(1:K), :);
end

%% EM

% 1. initialization
m = samples;
sigma = sigma_init * ones(K,1);
sigma_in = sigma;
sigma_out = sigma;
sigma_def = sigma;
p = 1/K * ones(K,1);


% iterations
for it = 1:max_it

	% 2. compute responsibilities - [ Expectation step ]
	gamma = zeros(N,K);
	dist = distance(data', m');
	for k=1:K
		gamma(:,k) = p(k) * (sigma(k))^(-D) * exp((-1/(2*sigma(k)^2)) * (dist(:,k).^2));
	end

	% normalize per point
	for n=1:N
		sg = sum(gamma(n,:));
		if(sg), gamma(n,:) = gamma(n,:) / sg;
		else
			[dummy, j] = min(dist(n,:));
			gamma(n,:) = zeros(1,K);
			gamma(n, min(j)) = 1;
		end
	end
	Nk = sum(gamma)';

	% display
	if (mod(it, n_it)==0),
		disp_gm(x, y, m, sigma, sigma_in, sigma_out, sigma_def, it, K);
	end

	% 3. update parameters - [ Maximization step ]
	gamma_max = max(gamma, [], 2);

	% 3.1 mixing coefficients and means
	p = Nk / N;
	for k=1:K
		m(k,:) = 0;
		for n=1:N,
			m(k,:) = m(k,:) + gamma(n,k) * data(n,:);
		end
		m(k,:) = (1/Nk(k)) * m(k,:);
	end

	% 3.2 expand - sigma
	for k=1:K
		Nin_k = 0;
		Nout_k = 0;
		sigma_in(k) = 0;
		sigma_out(k) = 0;
		for n=1:N
			if gamma(n,k) == gamma_max(n),
				Nin_k = Nin_k + gamma(n,k);
				sigma_in(k) = sigma_in(k) + gamma(n,k) * dist(n,k)^2;
			else
				Nout_k = Nout_k + gamma(n,k);
				sigma_out(k) = sigma_out(k) + gamma(n,k) * dist(n,k)^2;
			end
		end
		if Nin_k, sigma_in(k) = sqrt( (1/D) * (1/Nin_k) * sigma_in(k) ); end
		if Nout_k, sigma_out(k) = sqrt( (1/D) * (1/Nout_k) * sigma_out(k) ); end
		if isinf(sigma_out(k)), sigma_out(k) = 0; end % only for numerical stability
		in = (1 - lambda) * (Nin_k/Nk(k));
		sigma(k) = sqrt( sigma0^2 + in * sigma_in(k).^2 + (1-in) * sigma_out(k).^2 );
		in = Nin_k / Nk(k);
		sigma_def(k) = sqrt( sigma0^2 + in * sigma_in(k).^2 + (1-in) * sigma_out(k).^2 );
	end

	% 4. [ Purge step ]
	% 4.1 component responsibilities
	dist_k = distance(m', m');
	gamma_k = zeros(K);
	for k = 1:K
		sigma_j = sqrt(sigma.^2 + sigma(k)^2);
		gamma_k(:,k) = p(k) * (sigma_j).^(-D) .* exp((-1./(2*sigma_j.^2)).*(dist_k(:,k).^2));
	end
	for n = 1:K
		sg = sum( gamma_k(n,:) );
		if (sg), gamma_k(n,:) = gamma_k(n,:) / sg;
		else
			[dummy, j] = min(dist_k(n,:));
			gamma_k(n,:) = zeros(1,K);
			gamma_k(n, min(j)) = 1;
		end
	end

	% 4.2 component overlap & purging
	ik = 1;
	while ik <= K
		[dummy,id] = sort(p);
		id = flipud(id);
		k = id(ik);
		if ( gamma_k(k,k) / sum(gamma_k(k,id(1:ik))') < tau )
				p(k) = [];
				sigma(k) = [];
				sigma_in(k) = [];
				sigma_out(k) = [];
				sigma_def(k) = [];
				m(k,:) = [];
				K = K-1;
				gamma_k(:,k) = [];
				gamma_k(k,:) = [];
			else
				ik = ik+1;
			end
		end

end
